@@ -14,9 +14,8 @@
  * Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public
- * License along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
- * 02111-1307, USA.
+ * License along with this program; if not, see
+ * <http://www.gnu.org/licenses/>.
  */
 
 #include "config.h"
@@ -88,6 +87,16 @@ default_path (GkdSecretService *self)
 	gchar *old_directory;
 	gchar *new_directory;
 
+#if WITH_DEBUG
+	if (self->alias_directory == NULL) {
+		const gchar *path = g_getenv ("GNOME_KEYRING_TEST_PATH");
+		if (path && path[0]) {
+			self->alias_directory = g_strdup (path);
+			g_debug ("Alias directory was overridden by tests: %s", path);
+		}
+	}
+#endif
+
 	if (self->alias_directory == NULL) {
 		new_directory = g_build_filename (g_get_user_data_dir (), "keyrings", NULL);
 		old_directory = g_build_filename (g_get_home_dir (), ".gnome2", "keyrings", NULL);
